Object subclass: #TTContourConstruction
	instanceVariableNames: 'points '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Support'!
!TTContourConstruction commentStamp: '<historical>' prior: 0!
This class represents a temporary contour structure during the construction of a TTGlyph from a TrueType file.

Instance variables:
	points	<Array of: TTPoint>	The points defining this contour!


!TTContourConstruction methodsFor: 'accessing'!
points
	^points! !

!TTContourConstruction methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:34'!
points: anArray
	points _ anArray asArray.! !

!TTContourConstruction methodsFor: 'accessing'!
segments

	| segments |
	segments _ OrderedCollection new.
	self segmentsDo:[:seg| segments add: seg].
	^segments! !


!TTContourConstruction methodsFor: 'enumerating' stamp: 'hmm 10/28/2001 21:55'!
segmentsDo: aBlock
	"Evaluate aBlock with the segments of the receiver. This may either be straight line
	segments or quadratic bezier curves. The decision is made upon the type flags
	in TTPoint as follows:
	a) 	Two subsequent #OnCurve points define a straight segment
	b) 	An #OnCurve point followed by an #OffCurve point followed 
		by an #OnCurve point defines a quadratic bezier segment
	c)	Two subsequent #OffCurve points have an implicitely defined 
		#OnCurve point at half the distance between them"
	| last next mid index i |
	last _ points first.
	"Handle case where first point is off-curve"
	(last type == #OnCurve) ifFalse: [
		i _ points findFirst: [:pt | pt type == #OnCurve].
		i = 0
			ifTrue: [mid _ TTPoint new
							type: #OnCurve;
							x: points first x + points last x // 2;
							y: points first y + points last y // 2.
					points _ (Array with: mid), points]
			ifFalse: [points _ (points copyFrom: i to: points size), (points copyFrom: 1 to: i)].
		last _ points first].
	index _ 2.
	[index <= points size] whileTrue:[
		mid _ points at: index.
		mid type == #OnCurve ifTrue:[
			"Straight segment"
			aBlock value: (LineSegment from: last asPoint to: mid asPoint).
			last _ mid.
		] ifFalse:["Quadratic bezier"
			"Read ahead if the next point is on curve"
			next _ (index < points size) ifTrue:[points at: (index+1)] ifFalse:[points first].
			next type == #OnCurve ifTrue:[
				"We'll continue after the end point"
				index _ index + 1.
			] ifFalse:[ "Calculate center"
				next _ (next asPoint + mid asPoint) // 2].
			aBlock value:(Bezier2Segment from: last asPoint via: mid asPoint to: next asPoint).
			last _ next].
		index _ index + 1].
	(index = (points size + 1)) ifTrue:[
		aBlock value:(LineSegment from: points last asPoint to: points first asPoint)]! !


!TTContourConstruction methodsFor: 'converting' stamp: 'di 11/21/1999 20:19'!
asCompressedPoints
	"Return the receiver compressed into a PointArray.
	All lines will be converted into bezier segments with
	the control point set to the start point"
	| out minPt maxPt fullRange |
	minPt _ -16r7FFF asPoint.
	maxPt _ 16r8000 asPoint.
	"Check if we need full 32bit range"
	fullRange _ points anySatisfy: [:any| any asPoint < minPt or:[any asPoint > maxPt]].
	fullRange ifTrue:[
		out _ WriteStream on: (PointArray new: points size).
	] ifFalse:[
		out _ WriteStream on: (ShortPointArray new: points size).
	].
	self segmentsDo:[:segment|
		out nextPut: segment start.
		segment isBezier2Segment 
			ifTrue:[out nextPut: segment via]
			ifFalse:[out nextPut: segment start].
		out nextPut: segment end.
	].
	^out contents! !


!TTContourConstruction methodsFor: 'printing'!
printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPut:$(;
		print: points size;
		"space;
		print: self type;"
		nextPut:$)! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTContourConstruction class
	instanceVariableNames: ''!

!TTContourConstruction class methodsFor: 'instance creation'!
on: points

	^self new points: points! !


Object subclass: #TTFontReader
	instanceVariableNames: 'charMap glyphs nGlyphs kernPairs infoBar fontDescription '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Support'!

!TTFontReader methodsFor: 'public' stamp: 'ar 11/2/1998 23:51'!
readFrom: aStream

	| fontData headerEntry maxProfileEntry nameEntry indexLocEntry charMapEntry glyphEntry horzHeaderEntry horzMetricsEntry kerningEntry glyphOffset cmap numHMetrics indexToLocFormat |

	"Read the raw font byte data"
	(aStream respondsTo: #binary) ifTrue:[aStream binary].
	fontData _ aStream contents asByteArray.
	fontDescription _ TTFontDescription new.

	"Search the tables required to build the font"
	(headerEntry _ self getTableDirEntry: 'head' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a header table'].
	(maxProfileEntry _ self getTableDirEntry: 'maxp' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a maximum profile table'].
	(nameEntry _ self getTableDirEntry: 'name' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a name table'].
	(indexLocEntry _ self getTableDirEntry: 'loca' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a relocation table'].
	(charMapEntry _ self getTableDirEntry: 'cmap' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a character map table'].
	(glyphEntry _ self getTableDirEntry: 'glyf' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a glyph table'].
	(horzHeaderEntry _ self getTableDirEntry: 'hhea' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a horizontal header table'].
	(horzMetricsEntry _ self getTableDirEntry: 'hmtx' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a horizontal metrics table'].
	(kerningEntry _ self getTableDirEntry: 'kern' from: fontData) == nil ifTrue:[
		Transcript cr; show:'This font does not have a kerning table';endEntry].


	"Process the data"
	indexToLocFormat _ self processFontHeaderTable: headerEntry.
	self processMaximumProfileTable: maxProfileEntry.
	self processNamingTable: nameEntry.
	glyphOffset _ self processIndexToLocationTable: indexLocEntry format: indexToLocFormat.
	cmap _ self processCharacterMappingTable: charMapEntry.
	(cmap == nil or:[cmap value == nil])
		ifTrue:[^self error:'This font has no suitable character mappings'].
	self processGlyphDataTable: glyphEntry offsets: glyphOffset.
	numHMetrics _ self processHorizontalHeaderTable: horzHeaderEntry.
	self processHorizontalMetricsTable: horzMetricsEntry length: numHMetrics.
	kerningEntry isNil 
		ifTrue:[kernPairs _ #()]
		ifFalse:[self processKerningTable: kerningEntry].
	charMap _ self processCharMap: cmap.
	fontDescription setGlyphs: glyphs mapping: charMap.
	fontDescription setKernPairs: kernPairs.
	^fontDescription! !


!TTFontReader methodsFor: 'processing' stamp: 'sma 3/11/2000 20:30'!
processCharMap: assoc
	"Process the given character map"

	| charTable glyph cmap |
	cmap _ assoc value.
	charTable _ Array new: 256 withAll: glyphs first. "Initialize with default glyph"

	assoc key = 1 ifTrue: "Mac encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph _ glyphs at: (cmap at: i) + 1.
			charTable at: i put: glyph]].

	assoc key = 3 ifTrue: "Win encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph _ glyphs at: (cmap at: i) + 1.
			charTable at: (self winToMac: i) put: glyph]].

	^ charTable! !

!TTFontReader methodsFor: 'processing' stamp: 'sma 1/1/2000 17:01'!
processCharacterMappingTable: entry
	"Read the font's character to glyph index mapping table.
	If an appropriate mapping can be found then return an association
	with the format identifier and the contents of the table"
	| copy initialOffset nSubTables pID sID offset cmap assoc |
	initialOffset _ entry offset.
	entry skip: 2. "Skip table version"
	nSubTables _ entry nextUShort.
	1 to: nSubTables do:[:i|
		pID _ entry nextUShort.
		sID _ entry nextUShort.
		offset _ entry nextULong.
		"Check if this is either a Macintosh encoded table
		or a Windows encoded table"
		(pID = 1 or:[pID = 3]) ifTrue:[
			"Go to the beginning of the table"
			copy _ entry copy.
			copy offset: initialOffset + offset.
			cmap _ self decodeCmapFmtTable: copy.
			(pID = 1 and: [cmap notNil]) "Prefer Macintosh encoding over everything else"
				ifTrue: [^ pID -> cmap].
			assoc _ pID -> cmap. "Keep it in case we don't find a Mac encoded table"
		].
	].
	^assoc! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/15/1998 01:00'!
processCompositeGlyph: glyph contours: nContours from: entry
	"Read a composite glyph from the font data. The glyph passed into this method contains some state variables that must be copied into the resulting composite glyph."
	| flags glyphIndex hasInstr cGlyph ofsX ofsY iLen a11 a12 a21 a22 m |
	cGlyph _ TTCompositeGlyph new.
	a11 _ a22 _ 16r4000.	"1.0 in F2Dot14"
	a21 _ a12 _ 0.		"0.0 in F2Dot14"
	"Copy state"
	cGlyph bounds: glyph bounds; glyphIndex: glyph glyphIndex.
	hasInstr _ false.
	[ flags _ entry nextUShort.
	glyphIndex _ entry nextUShort + 1.
	(flags bitAnd: 1) = 1 ifTrue:[
		ofsX _ entry nextShort.
		ofsY _ entry nextShort.
	] ifFalse:[
		(ofsX _ entry nextByte) > 127 ifTrue:[ofsX _ ofsX - 256].
		(ofsY _ entry nextByte) > 127 ifTrue:[ofsY _ ofsY - 256]].
	((flags bitAnd: 2) = 2) ifFalse:[self halt].
	(flags bitAnd: 8) = 8 ifTrue:[
		a11 _ a22 _ entry nextShort].
	(flags bitAnd: 64) = 64 ifTrue:[
		a11 _ entry nextShort.
		a22 _ entry nextShort].
	(flags bitAnd: 128) = 128 ifTrue:[
		"2x2 transformation"
		a11 _ entry nextShort.
		a21 _ entry nextShort.
		a12 _ entry nextShort.
		a22 _ entry nextShort].
	m _ MatrixTransform2x3 new.
	"Convert entries from F2Dot14 to float"
	m a11: (a11 asFloat / 16r4000).
	m a12: (a12 asFloat / 16r4000).
	m a21: (a21 asFloat / 16r4000).
	m a22: (a22 asFloat / 16r4000).
	m a13: ofsX.
	m a23: ofsY.
	cGlyph addGlyph: (glyphs at: glyphIndex) transformation: m.
	hasInstr _ hasInstr or:[ (flags bitAnd: 256) = 256].
	"Continue as long as the MORE_COMPONENTS bit is set"
	(flags bitAnd: 32) = 32] whileTrue.
	hasInstr ifTrue:[
		iLen _ entry nextUShort.
		entry skip: iLen].
	^cGlyph! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:42'!
processFontHeaderTable: entry
"Value				Data Type    Description
unitsPerEm			USHORT      Granularity of the font's em square.
xMax				USHORT      Maximum X-coordinate for the entire font.
xMin				USHORT      Minimum X-coordinate for the entire font.
yMax				USHORT      Maximum Y-coordinate for the entire font.
yMin				USHORT      Minimum Y-coordinate for the entire font.
indexToLocFormat	SHORT       Used when processing the Index To Loc Table."
	| origin corner units indexToLocFormat |
	entry skip: 4. "Skip table version number"
	entry skip: 4. "Skip font revision number"
	entry skip: 4. "Skip check sum adjustment"
	entry skip: 4. "Skip magic number"
	entry skip: 2. "Skip flags"

	units _ entry nextUShort.

	entry skip: 8. "Skip creation date"
	entry skip: 8. "Skip modification date"

	"Get min/max values of all glyphs"
	origin _ entry nextShort @ entry nextShort.
	corner _ entry nextShort @ entry nextShort.

	entry skip: 2. "Skip mac style"
	entry skip: 2. "Skip lowest rec PPEM"
	entry skip: 2. "Skip font direction hint"
	indexToLocFormat _ entry nextShort.

	fontDescription setBounds: (origin corner: corner) unitsPerEm: units.
	^indexToLocFormat! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/3/1998 14:43'!
processGlyphDataTable: entry offsets: offsetArray
	"Read the actual glyph data from the font.
	offsetArray contains the start offsets in the data for each glyph."
	| initialOffset glyph nextOffset glyphLength glyphOffset nContours origin corner |
	initialOffset _ entry offset.
	glyphs _ Array new: nGlyphs.
	1 to: nGlyphs do:[:i | 
		glyphs at: i put: (TTGlyph new glyphIndex: i-1)].
	'Reading glyph data' 
		displayProgressAt: Sensor cursorPoint
		from: 1 to: nGlyphs during:[:bar|

	1 to: nGlyphs do:[:glyphIndex |
		bar value: glyphIndex.
		glyph _ glyphs at: glyphIndex.
		glyphOffset _ offsetArray at: glyphIndex.
		nextOffset _ offsetArray at: glyphIndex+1.
		glyphLength _ nextOffset - glyphOffset.
		glyphLength = 0 ifFalse:[
			entry offset: initialOffset + glyphOffset.
			nContours _ entry nextShort.
			origin _ entry nextShort @ entry nextShort.
			corner _ entry nextShort @ entry nextShort.
			glyph bounds: (origin corner: corner).
			nContours >= 0 ifTrue:[
				self processSimpleGlyph: glyph contours: nContours from: entry
			] ifFalse:[
				glyph _ self processCompositeGlyph: glyph contours: nContours from: entry.
				glyphs at: glyphIndex put: glyph]]]
	].! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:40'!
processHorizontalHeaderTable: entry
"
ascender           SHORT          Typographic ascent.
descender          SHORT          Typographic descent.
lineGap            SHORT          Typographic lineGap.
numberOfHMetrics   USHORT         Number hMetric entries in the HTMX
                                               Table; may be smaller than the total
                                             number of glyphs.
"
	| asc desc lGap numHMetrics |
	entry skip: 4. "Skip table version"
	asc _ entry nextShort.
	desc _ entry nextShort.
	lGap _ entry nextShort.
	entry skip: 2. "Skip advanceWidthMax"
	entry skip: 2. "Skip minLeftSideBearing"
	entry skip: 2. "Skip minRightSideBearing"
	entry skip: 2. "Skip xMaxExtent"
	entry skip: 2. "Skip caretSlopeRise"
	entry skip: 2. "Skip caretSlopeRun"
	entry skip: 10. "Skip 5 reserved shorts"
	entry skip: 2. "Skip metricDataFormat"

	numHMetrics _ entry nextUShort.

	fontDescription setAscender: asc descender: desc lineGap: lGap.
	^numHMetrics! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:40'!
processHorizontalMetricsTable: entry length: numHMetrics
	"Extract the advance width, left side bearing, and right
	side bearing for each glyph from the Horizontal Metrics Table."
	|  index lastAW glyph |
	index _ 1.
	[index <= numHMetrics] whileTrue:[
		glyph _ glyphs at: index.
		glyph advanceWidth: entry nextUShort.
		glyph leftSideBearing: entry nextShort.
		glyph updateRightSideBearing.
		index _ index + 1].
	index = (nGlyphs +1) ifTrue:[^true].
	lastAW _ (glyphs at: index-1) advanceWidth.

	[index <= nGlyphs] whileTrue:[
		glyph _ glyphs at: index.
		glyph advanceWidth: lastAW.
		glyph leftSideBearing: entry nextShort.
		glyph updateRightSideBearing.
		index _ index + 1].! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:43'!
processIndexToLocationTable: entry format: indexToLocFormat
"glyphOffset    ULONG[numGlyphs]   An array that contains each glyph's
                                 offset into the Glyph Data Table.
"	| glyphOffset offset|
	glyphOffset _ Array new: nGlyphs+1.
	1 to: nGlyphs+1 do:[:i|
		(indexToLocFormat = 0) ifTrue:[ "Format0: offset/2 is stored"
			offset _ entry nextUShort * 2.
		] ifFalse:["Format1: store actual offset"
			offset _ entry nextULong].
		glyphOffset at: i put: offset].
	^glyphOffset! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/1/1998 23:21'!
processKerningTable: entry
	"Extract the kerning information for pairs of glyphs."
	| covLow covHigh nKernPairs kp |
	entry skip: 2. "Skip table version"
	entry skip: 2. "Skip number of sub tables -- we're using the first one only"
	entry skip: 2. "Skip current subtable number"
	entry skip: 2. "Skip length of subtable"
	covHigh _ entry nextByte.
	covLow _ entry nextByte.

	"Make sure the format is right (kerning table and format type 0)"
	((covLow bitAnd: 2) = 2 or:[ covHigh ~= 0]) ifTrue:[^false].
	nKernPairs _ entry nextUShort.
	entry skip: 2. "Skip search range"
	entry skip: 2. "Skip entry selector"
	entry skip: 2. "Skip range shift"
	kernPairs _ Array new: nKernPairs.
	1 to: nKernPairs do:[:i|
		kp _ TTKernPair new.
		kp left: entry nextUShort.
		kp right: entry nextUShort.
		kp value: entry nextShort.
		kernPairs at: i put: kp].
	^true! !

!TTFontReader methodsFor: 'processing'!
processMaximumProfileTable: entry
"
numGlyphs         USHORT      The number of glyphs in the font.
"
	entry skip: 4. "Skip Table version number"
	nGlyphs _ entry nextUShort.! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:38'!
processNamingTable: entry
"copyright         CHARPTR     The font's copyright notice.
familyName        CHARPTR     The font's family name.
subfamilyName     CHARPTR     The font's subfamily name.
uniqueName        CHARPTR     A unique identifier for this font.
fullName          CHARPTR     The font's full name (a combination of
                                          familyName and subfamilyName).
versionName       CHARPTR     The font's version string.
"	| nRecords initialOffset storageOffset pID sID lID nID length offset multiBytes string strings |
	strings _ Array new: 8.
	strings atAllPut:''.
	initialOffset _ entry offset.
	entry skip: 2. "Skip format selector"
	"Get the number of name records"
	nRecords _ entry nextUShort.
	"Offset from the beginning of this table"
	storageOffset _ entry nextUShort + initialOffset.
	1 to: nRecords do:[:i|
		pID _ entry nextUShort.
		sID _ entry nextUShort.
		lID _ entry nextUShort.
		nID _ entry nextUShort.
		length _ entry nextUShort.
		offset _ entry nextUShort.
		"Read only Macintosh or Microsoft strings"
		(pID = 1 or:[pID = 3 and:[sID = 1]]) ifTrue:[
			"MS uses Unicode all others single byte"
			multiBytes _ pID = 3.
			string _ entry stringAt: storageOffset + offset length: length multiByte: multiBytes.
			"Put the name at the right location.
			Note: We prefer Macintosh strings about everything else."
			nID < strings size ifTrue:[
				(pID = 1 or:[(strings at: nID+1) = ''])
					ifTrue:[strings at: nID+1 put: string].
			].
		].
	].
	fontDescription setStrings: strings.! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/1/1998 22:18'!
processSimpleGlyph: glyph contours: nContours from: entry

	| endPts  nPts iLength flags |
	endPts _ Array new: nContours.
	1 to: nContours do:[:i| endPts at: i put: entry nextUShort].
	glyph initializeContours: nContours with: endPts.
	nPts _ endPts last + 1.
	iLength _ entry nextUShort. "instruction length"
	entry skip: iLength.
	flags _ self getGlyphFlagsFrom: entry size: nPts.
	self readGlyphXCoords: entry glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	self readGlyphYCoords: entry glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	glyph buildContours.! !


!TTFontReader methodsFor: 'private' stamp: 'sma 1/1/2000 19:17'!
decodeCmapFmtTable: entry
	| cmapFmt length cmap firstCode entryCount segCount segments offset code |
	cmapFmt _ entry nextUShort.
	length _ entry nextUShort.
	entry skip: 2. "skip version"

	cmapFmt = 0 ifTrue: "byte encoded table"
		[length _ length - 6. 		"should be always 256"
		length <= 0 ifTrue: [^ nil].	"but sometimes, this table is empty"
		cmap _ Array new: length.
		entry nextBytes: length into: cmap startingAt: entry offset.
		^ cmap].

	cmapFmt = 4 ifTrue: "segment mapping to deltavalues"
		[segCount _ entry nextUShort // 2.
		entry skip: 6. "skip searchRange, entrySelector, rangeShift"
		segments _ Array new: segCount.
		segments _ (1 to: segCount) collect: [:e | Array new: 4].
		1 to: segCount do: [:i | (segments at: i) at: 2 put: entry nextUShort]. "endCount"
		entry skip: 2. "skip reservedPad"
		1 to: segCount do: [:i | (segments at: i) at: 1 put: entry nextUShort]. "startCount"
		1 to: segCount do: [:i | (segments at: i) at: 3 put: entry nextShort]. "idDelta"
		offset _ entry offset.
		1 to: segCount do: [:i | (segments at: i) at: 4 put: entry nextUShort]. "idRangeOffset"
		cmap _ Array new: 256 withAll: 0. "could be larger, but Squeak can't handle that"
		segments withIndexDo:
			[:seg :si |
			seg first to: seg second do:
				[:i |
				i < 256 ifTrue:
					[seg last > 0 ifTrue:
						["offset to glypthIdArray - this is really C-magic!!"
						entry offset: i - seg first - 1 * 2 + seg last + si + si + offset. 
						code _ entry nextUShort.
						code > 0 ifTrue: [code _ code + seg third]]
					ifFalse:
						["simple offset"
						code _ i + seg third].
					cmap at: i + 1 put: code]]].
		^ cmap].

	cmapFmt = 6 ifTrue: "trimmed table"
		[firstCode _ entry nextUShort.
		entryCount _ entry nextUShort.
		cmap _ Array new: entryCount + firstCode withAll: 0.
		entryCount timesRepeat:
			[cmap at: (firstCode _ firstCode + 1) put: entry nextUShort].
		^ cmap].
	^ nil! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:33'!
getGlyphFlagsFrom: entry size: nPts
	"Read in the flags for this glyph.  The outer loop gathers the flags that
	are actually contained in the table.  If the repeat bit is set in a flag
	then the next byte is read from the table; this is the number of times
	to repeat the last flag.  The inner loop does this, incrementing the
	outer loops index each time."
	| flags index repCount flagBits |
	flags _ ByteArray new: nPts.
	index _ 1.
	[index <= nPts] whileTrue:[
		flagBits _ entry nextByte.
		flags at: index put: flagBits.
		(flagBits bitAnd: 8) = 8 ifTrue:[
			repCount _ entry nextByte.
			repCount timesRepeat:[
				index _ index + 1.
				flags at: index put: flagBits]].
		index _ index + 1].
	^flags! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:33'!
getTableDirEntry: tagString from: fontData
	"Find the table named tagString in fontData and return a table directory entry for it."
	| nTables pos currentTag tag |
	nTables _ fontData shortAt: 5 bigEndian: true.
	tag _ ByteArray new: 4.
	1 to: 4 do:[:i| tag byteAt: i put: (tagString at: i) asInteger].
	tag _ tag longAt: 1 bigEndian: true.
	pos _ 13.
	1 to: nTables do:[:i|
		currentTag _ fontData longAt: pos bigEndian: true.
		currentTag = tag ifTrue:[^TTFontTableDirEntry on: fontData at: pos].
		pos _ pos+16].
	^nil! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:36'!
readGlyphXCoords:entry glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the x coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits xValue contour ttPoint |
	startPoint _ 1.
	1 to: nContours do:[:i|
		contour _ glyph contours at: i.
		"Get the end point"
		endPoint _ (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint _ contour points at: (j - startPoint + 1).
			flagBits _ flags at: j.
			"If bit zero in the flag is set then this point is an on-curve
			point, if not, then it is an off-curve point."
			(flagBits bitAnd: 1) = 1 
				ifTrue:[ ttPoint type: #OnCurve]
				ifFalse:[ttPoint type: #OffCurve].
			"First we check to see if bit one is set.  This would indicate that
			the corresponding coordinate data in the table is 1 byte long.
			If the bit is not set, then the coordinate data is 2 bytes long."
			(flagBits bitAnd: 2) = 2 ifTrue:[ "one byte"
				xValue _ entry nextByte.
				xValue _ (flagBits bitAnd: 16)=16 ifTrue:[xValue] ifFalse:[xValue negated].
				ttPoint x: xValue.
			] ifFalse:[ "two byte"
				"If bit four is set, then this coordinate is the same as the
				last one, so the relative offset (of zero) is stored.  If bit
				is not set, then read in two bytes and store it as a signed value."
				(flagBits bitAnd: 16) = 16 ifTrue:[ ttPoint x: 0 ]
				ifFalse:[
					xValue _ entry nextShort.
					ttPoint x: xValue]]].
		startPoint _ endPoint + 1]! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:37'!
readGlyphYCoords:entry glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the y coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits yValue contour ttPoint |
	startPoint _ 1.
	1 to: nContours do:[:i|
		contour _ glyph contours at: i.
		"Get the end point"
		endPoint _ (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint _ contour points at: (j - startPoint + 1).
			flagBits _ flags at: j.
			"Check if this value one or two byte encoded"
			(flagBits bitAnd: 4) = 4 ifTrue:[ "one byte"
				yValue _ entry nextByte.
				yValue _ (flagBits bitAnd: 32)=32 ifTrue:[yValue] ifFalse:[yValue negated].
				ttPoint y: yValue.
			] ifFalse:[ "two byte"
				(flagBits bitAnd: 32) = 32 ifTrue:[ ttPoint y: 0 ]
				ifFalse:[
					yValue _ entry nextShort.
					ttPoint y: yValue]]].
		startPoint _ endPoint + 1]! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/1/1998 21:01'!
warn: aString
	Transcript cr; show: aString; endEntry.! !

!TTFontReader methodsFor: 'private' stamp: 'sma 3/11/2000 20:33'!
winToMac: index
	^ (index - 1) asCharacter isoToSqueak asciiValue + 1! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTFontReader class
	instanceVariableNames: ''!

!TTFontReader class methodsFor: 'instance creation' stamp: 'sma 5/5/2000 14:22'!
installTTF: ttfFileName asTextStyle: textStyleName sizes: sizeArray
	"TTFontReader
		installTTF: 'F:\fonts\amazon__.TTF' 
		asTextStyle: #Amazon
		sizes: #(24 60)"

	| ttf fontArray |
	ttf _ self parseFileNamed: ttfFileName.
	fontArray _ sizeArray collect:
		[:each |
		(ttf asStrikeFontScale: each / ttf unitsPerEm)
			name: textStyleName;
			pointSize: each].
	TextConstants at: textStyleName asSymbol put: (TextStyle fontArray: fontArray)! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 23:43'!
parseFileNamed: aString
	"TTFontReader parseFileNamed:'c:\windows\arial.ttf'"
	"TTFontReader parseFileNamed:'c:\windows\times.ttf'"
	| contents |
	contents _ (FileStream readOnlyFileNamed: aString) binary contentsOfEntireFile.
	^self readFrom: (ReadStream on: contents)! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 00:53'!
readFrom: aStream

	^self new readFrom: aStream! !


!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/6/2002 21:37'!
fileReaderServicesForFile: fullName suffix: suffix


	^(suffix = 'fnt')
		ifTrue: [
			self services]
		ifFalse: [#()]
! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/1/2002 19:31'!
openTTFFile: fullName 

	(TTFontReader parseFileNamed: fullName) asMorph open! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/1/2002 22:02'!
serviceOpenTrueTypeFont

	^ SimpleServiceEntry 
				provider: self 
				label: 'open true type font'
				selector: #openTTFFile:
				description: 'open true type font'! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/1/2002 22:03'!
services

	^ Array with: self serviceOpenTrueTypeFont
! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/1/2002 19:29'!
unload

	FileList unregisterFileReader: self ! !


Object subclass: #TTFontTableDirEntry
	instanceVariableNames: 'tag fontData offset length checkSum '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Support'!

!TTFontTableDirEntry methodsFor: 'initialize-release'!
on: fd at: index

	fontData _ fd.
	tag _ fontData longAt: index bigEndian: true.
	checkSum _ fontData longAt: index+4 bigEndian: true.
	offset _ (fontData longAt: index+8 bigEndian: true) + 1.
	length _ fontData longAt: index+12 bigEndian: true.! !


!TTFontTableDirEntry methodsFor: 'accessing'!
nextByte

	| value |
	value _ fontData byteAt: offset.
	offset _ offset + 1.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing'!
nextBytes: numBytes into: array startingAt: byteOffset

	1 to: numBytes do:[:i|
		array at: i put: (fontData byteAt: byteOffset + i - 1)].! !

!TTFontTableDirEntry methodsFor: 'accessing'!
nextLong

	| value |
	value _ fontData longAt: offset bigEndian: true.
	offset _ offset + 4.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing'!
nextShort

	| value |
	value _ fontData shortAt: offset bigEndian: true.
	offset _ offset + 2.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing'!
nextULong

	| value |
	value _ fontData unsignedLongAt: offset bigEndian: true.
	offset _ offset + 4.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing'!
nextUShort

	| value |
	value _ fontData unsignedShortAt: offset bigEndian: true.
	offset _ offset + 2.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing'!
offset
	^offset! !

!TTFontTableDirEntry methodsFor: 'accessing'!
offset: newOffset
	offset _ newOffset! !

!TTFontTableDirEntry methodsFor: 'accessing'!
skip: n
	"Skip n bytes"
	offset _ offset + n.! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: 'ar 11/1/1998 23:37'!
stringAt: stringOffset length: byteLength multiByte: aBoolean

	| string index stringLength |
	aBoolean ifFalse:[
		stringLength _ byteLength.
		string _ String new: stringLength.
		index _ stringOffset.
		1 to: stringLength do:[:i|
			string at: i put: (Character value: (fontData byteAt: index + i - 1))].
		^string
	] ifTrue:[
		stringLength _ byteLength // 2.
		string _ String new: stringLength.
		index _ stringOffset.
		1 to: stringLength do:[:i|
			string at: i put: (Character value: (fontData byteAt: index + 1)).
			index _ index + 2].
		^string]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTFontTableDirEntry class
	instanceVariableNames: ''!

!TTFontTableDirEntry class methodsFor: 'instance creation'!
on: fontData at: index

	^self new on: fontData at: index! !


Object subclass: #TTPoint
	instanceVariableNames: 'x y type '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Support'!

!TTPoint methodsFor: 'accessing'!
type
	^type! !

!TTPoint methodsFor: 'accessing'!
type: aSymbol

	type _ aSymbol! !

!TTPoint methodsFor: 'accessing'!
x
	^x! !

!TTPoint methodsFor: 'accessing'!
x: aNumber

	x _ aNumber! !

!TTPoint methodsFor: 'accessing'!
y
	^y! !

!TTPoint methodsFor: 'accessing'!
y: aNumber
	y _ aNumber! !


!TTPoint methodsFor: 'printing'!
printOn: aStream

	aStream 
		nextPutAll: self class name;
		nextPut:$(;
		print: x;
		nextPut:$@;
		print: y;
		nextPut:$|;
		print: type;
		nextPut:$)! !


!TTPoint methodsFor: 'converting'!
asPoint
	^x@y! !


BorderedMorph subclass: #TTSampleFontMorph
	instanceVariableNames: 'font transform smoothing '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Support'!

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'ar 11/14/1998 22:52'!
doesBevels
	^false! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'sma 1/1/2000 18:08'!
font
	^ font! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'ar 11/15/1998 00:31'!
font: aTTFontDescription
	| morph |
	font _ aTTFontDescription.
	morph _ (TTSampleStringMorph font: font).
	morph extent: morph extent * 2.
	morph color: Color magenta.
	self addMorphCentered: morph.
	morph position: morph position x @ (self bounds bottom + 10).
	self privateFullMoveBy: self fullBounds origin negated.! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'sma 1/1/2000 17:53'!
smoothing
	^ smoothing! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'bf 10/18/1999 16:19'!
smoothing: aNumber
	smoothing _ aNumber.
	self changed! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'ar 11/14/1998 22:53'!
transform
	^transform ifNil:[self computeTransform].! !


!TTSampleFontMorph methodsFor: 'copying' stamp: 'sma 2/26/2000 19:20'!
veryDeepFixupWith: deepCopier
	"If fields were weakly copied, fix them here. If they were in the 
	tree being copied, fix them up, otherwise point to the originals!!!!"

	super veryDeepFixupWith: deepCopier.
	font _ deepCopier references at: font ifAbsent: [font]! !

!TTSampleFontMorph methodsFor: 'copying' stamp: 'sma 2/26/2000 19:18'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all,
	but shared. Warning!!!! Every instance variable defined in this class
	must be handled.  We must also implement veryDeepFixupWith:.
	See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	"font _ font"
	transform _ transform veryDeepCopyWith: deepCopier.
	smoothing _ smoothing veryDeepCopyWith: deepCopier! !


!TTSampleFontMorph methodsFor: 'drawing' stamp: 'ar 11/14/1998 23:12'!
areasRemainingToFill: aRectangle
	^ Array with: aRectangle! !

!TTSampleFontMorph methodsFor: 'drawing' stamp: 'sma 1/1/2000 17:56'!
drawCharactersOn: aCanvas
	| glyph origin r offset cy m |
	0 to: 255 do:[:i|
		glyph _ font at: i.
		origin _ font bounds extent * ((i \\ 16) @ (i // 16)).
		r _ origin extent: font bounds extent.
		offset _ r center - glyph bounds center.
		cy _ glyph bounds center y.
		m _ MatrixTransform2x3 withOffset: 0@cy.
		m _ m composedWithLocal: (MatrixTransform2x3 withScale: 1@-1).
		m _ m composedWithLocal: (MatrixTransform2x3 withOffset: 0@cy negated).
		m _ m composedWithGlobal: (MatrixTransform2x3 withOffset: offset).
		aCanvas asBalloonCanvas preserveStateDuring:[:balloonCanvas|
			balloonCanvas transformBy: m.
			balloonCanvas drawGeneralBezierShape: glyph contours
					color: color
					borderWidth: 0
					borderColor: Color black.
		].
	].! !

!TTSampleFontMorph methodsFor: 'drawing' stamp: 'ar 12/30/1998 10:49'!
drawOn: aCanvas
	| origin extent offset |
	(font isNil) 
		ifTrue:[^aCanvas frameRectangle: bounds color: Color black].
	origin _ self position asIntegerPoint.
	extent _ self extent asIntegerPoint.
	0 to: 16 do:[:i|
		offset _ (extent x * i // 16) @ (extent y * i // 16).
		aCanvas line: origin x @ (origin y + offset y) 
				to: (origin x + extent x) @ (origin y + offset y)
				width: borderWidth color: borderColor.
		aCanvas line: (origin x + offset x) @ origin y 
				to: (origin x + offset x) @ (origin y + extent y)
				width: borderWidth color: borderColor.
	].
	aCanvas asBalloonCanvas preserveStateDuring:[:balloonCanvas|
		balloonCanvas transformBy: self transform.
		balloonCanvas aaLevel: self smoothing.
		self drawCharactersOn: balloonCanvas.
	].! !


!TTSampleFontMorph methodsFor: 'geometry' stamp: 'ar 11/14/1998 22:53'!
extent: extentPoint
	super extent: extentPoint.
	transform _ nil.! !

!TTSampleFontMorph methodsFor: 'geometry' stamp: 'ar 11/14/1998 22:53'!
position: pos
	super position: pos.
	transform _ nil.! !


!TTSampleFontMorph methodsFor: 'halos and balloon help' stamp: 'sw 1/27/2000 15:43'!
addOptionalHandlesTo: aHalo box: box
	aHalo addHandleAt: box center color: Color magenta icon: nil on: #mouseDown send: #createSample to: self.! !

!TTSampleFontMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/14/1998 23:39'!
balloonHelpTextForHandle: aHandle
	aHandle eventHandler firstMouseSelector == #createSample
		ifTrue:[^'Create a sample string'].
	^super balloonHelpTextForHandle: aHandle! !


!TTSampleFontMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:40'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color black! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 18:43'!
initialize
	"initialize the state of the receiver"
	super initialize.
	smoothing _ 4.
	self extent: 300 @ 300! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'sma 4/30/2000 10:51'!
openInWorld
	Smalltalk isMorphic ifFalse: [^ self openInMVC].
	HandMorph attach: self! !


!TTSampleFontMorph methodsFor: 'initialize' stamp: 'ar 11/14/1998 23:50'!
open
	Smalltalk isMorphic 
		ifTrue:[self openInWorld]
		ifFalse:[self openInMVC]! !


!TTSampleFontMorph methodsFor: 'menu' stamp: 'ar 11/14/1998 23:46'!
createSample
	self world primaryHand attachMorph: (TTSampleStringMorph font: font)! !

!TTSampleFontMorph methodsFor: 'menu' stamp: 'ar 6/16/1999 07:20'!
getSmoothingLevel
	"Menu support"
	smoothing = 1 ifTrue:[^'turn on smoothing'].
	smoothing = 2 ifTrue:[^'more smoothing'].
	smoothing = 4 ifTrue:[^'turn off smoothing'].
! !

!TTSampleFontMorph methodsFor: 'menu' stamp: 'sma 1/1/2000 17:51'!
nextSmoothingLevel
	smoothing = 1
		ifTrue: [smoothing _ 2]
		ifFalse: [smoothing = 2
			ifTrue: [smoothing _ 4]
			ifFalse: [smoothing = 4
				ifTrue: [smoothing _ 1]]].
	self changed! !


!TTSampleFontMorph methodsFor: 'menus' stamp: 'ar 6/16/1999 07:21'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu addUpdating: #getSmoothingLevel action: #nextSmoothingLevel.! !


!TTSampleFontMorph methodsFor: 'rotate scale and flex' stamp: 'ar 11/15/1998 22:42'!
newTransformationMorph
	^MatrixTransformMorph new! !


!TTSampleFontMorph methodsFor: 'testing' stamp: 'ar 8/25/2001 19:09'!
canDrawBorder: aBorderStyle
	^aBorderStyle style == #simple! !


!TTSampleFontMorph methodsFor: 'updating' stamp: 'sma 1/1/2000 17:59'!
changed
	self invalidRect: (self fullBounds expandBy: 1)! !


!TTSampleFontMorph methodsFor: 'private' stamp: 'ar 11/15/1998 22:48'!
computeTransform
	| fullExtent scale |
	fullExtent _ font bounds extent * 16.
	scale _ self extent asFloatPoint / fullExtent asFloatPoint.
	transform _ MatrixTransform2x3 withScale: scale.
	transform _ transform composedWithGlobal: (MatrixTransform2x3 withOffset: self position).
	^transform! !

!TTSampleFontMorph methodsFor: 'private' stamp: 'ar 11/14/1998 22:55'!
privateMoveBy: delta
	super privateMoveBy: delta.
	transform _ nil.! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTSampleFontMorph class
	instanceVariableNames: ''!

!TTSampleFontMorph class methodsFor: 'instance creation' stamp: 'ar 11/14/1998 23:06'!
font: aTTFontDescription
	^self new font: aTTFontDescription! !


TTSampleFontMorph subclass: #TTSampleStringMorph
	instanceVariableNames: 'string ttBounds '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Support'!
!TTSampleStringMorph commentStamp: '<historical>' prior: 0!
I allow the display of a string in a TrueType font as a stand-alone morph.

Morph's color changes the inside of the characters.
Morph's borderColor changes the outline.

Many free fonts are stored at www.FontGuy.com.  
Use a normal web browser (not our Scamper) and go there.  
Choose 'categories' and browse to a font you like.  
Hold the mouse down on the example text in that font.  
When the menu comes up, choose "Copy this link location".  
Come back into Squeak, choose "load font from web..."
from my menu, and paste in the url.!


!TTSampleStringMorph methodsFor: 'accessing' stamp: 'tk 12/10/2001 16:21'!
font: aTTFontDescription
	font _ aTTFontDescription.
	string ifNil: [self string: aTTFontDescription fullName]
		ifNotNil: [self initializeString].! !

!TTSampleStringMorph methodsFor: 'accessing' stamp: 'sma 1/1/2000 18:08'!
string
	^ string! !

!TTSampleStringMorph methodsFor: 'accessing' stamp: 'ar 11/14/1998 23:53'!
string: aString
	string _ aString.
	self initializeString.! !


!TTSampleStringMorph methodsFor: 'drawing' stamp: 'ar 12/30/1998 10:51'!
drawOn: aCanvas
	| xStart glyph |
	(font isNil or:[string isNil or:[string isEmpty]]) 
		ifTrue:[^aCanvas frameRectangle: bounds color: Color black].
	xStart _ 0.
	aCanvas asBalloonCanvas preserveStateDuring:[:balloonCanvas|
		balloonCanvas transformBy: self transform.
		balloonCanvas aaLevel: self smoothing.
		string do:[:char|
			glyph _ font at: char.
			balloonCanvas preserveStateDuring:[:subCanvas|
				subCanvas transformBy: (MatrixTransform2x3 withOffset: xStart@0).
				subCanvas 
					drawGeneralBezierShape: glyph contours
					color: color 
					borderWidth: borderWidth 
					borderColor: borderColor].
			xStart _ xStart + glyph advanceWidth.
		].
	].! !


!TTSampleStringMorph methodsFor: 'geometry testing' stamp: 'dgd 2/22/2003 14:42'!
containsPoint: aPoint 
	"^ super containsPoint: aPoint"

	"so much faster..."

	| picker |
	(self bounds containsPoint: aPoint) ifFalse: [^false].
	picker := BalloonCanvas on: (Form extent: 1 @ 1 depth: 32).
	picker transformBy: (MatrixTransform2x3 withOffset: aPoint negated).
	self drawOn: picker.
	^(picker form bits first) ~= 0! !


!TTSampleStringMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/14/1998 23:44'!
addOptionalHandlesTo: aHalo box: box! !


!TTSampleStringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:40'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 0! !

!TTSampleStringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ {Color magenta. Color yellow. Color orange. Color lightGray} atRandom! !


!TTSampleStringMorph methodsFor: 'initialize' stamp: 'sma 1/1/2000 18:08'!
initializeString
	| xStart char glyph |
	(font isNil or: [string isNil]) ifTrue: [^ self].
	xStart _ 0.
	ttBounds _ 0@0 corner: 0@0.
	1 to: string size do:
		[:i |
		char _ string at: i.
		glyph _ font at: char.
		ttBounds _ ttBounds quickMerge: (glyph bounds translateBy: xStart@0).
		xStart _ xStart + glyph advanceWidth.
	].
	self extent: ttBounds extent // 40.
	borderWidth _ ttBounds height // 40! !


!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/7/2001 18:03'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'edit contents...' action: #edit.
	aCustomMenu add: 'how to find more fonts...' action: #howTo.
	aCustomMenu add: 'load font from web...' action: #loadFromURL.! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/7/2001 11:24'!
edit
	"Allow the user to change the text in a crude way"

	| str |
	str _ FillInTheBlankMorph request: 'Type in new text for this TrueType displayer.'
				 initialAnswer: 'some text'.
	str isEmpty ifTrue: [^ self].
	self string: str.
! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/10/2001 16:02'!
howTo

	self inform: 'Many free fonts are stored at www.FontGuy.com.  
Use a normal web browser (not our Scamper) and go there.  
Choose ''categories'' and browse to a font you like.  
Hold the mouse down on the example text in that font.  
When the menu comes up, choose "Copy this link location".  
Come back into Squeak, choose "load font from web..."
from this menu, and paste in the url.'! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/10/2001 16:03'!
loadFromURL
	"Allow the user to change the text in a crude way"

	| url |
	url _ FillInTheBlankMorph request: ' Type in the url for a TrueType font on the web. '
				 initialAnswer: 'http://www.fontguy.com/download.asp?fontid=1494'.
	url isEmpty ifTrue: [^ self].
	self loadFromURL: url.
! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/10/2001 16:03'!
loadFromURL: urlString
	"Fetch the file, unarchive, unzip, and use as my font."

	| rawStrm |
	rawStrm _ HTTPSocket httpGet: urlString. 	"Later use an HttpURL?"
	self font: (TTFontReader readFrom: rawStrm asUnZippedStream).
! !


!TTSampleStringMorph methodsFor: 'parts bin' stamp: 'tk 12/10/2001 17:36'!
initializeToStandAlone
	"Make me into an example"

	| dd |
	dd _ TTFontDescription default.
	dd ifNil: [^ RectangleMorph initializeToStandAlone].	"not available"

	super initializeToStandAlone.
	self font: dd; color: (TranslucentColor r: 1.0 g: 0.097 b: 1.0 alpha: 0.6).
	self string: 'TrueType fonts are beautiful'.
! !


!TTSampleStringMorph methodsFor: 'private' stamp: 'ar 11/14/1998 22:04'!
computeTransform
	| cy |
	cy _ bounds origin y + bounds corner y * 0.5.
	transform _ MatrixTransform2x3 
			transformFromLocal: (ttBounds insetBy: borderWidth negated)
			toGlobal: bounds.
	transform _ transform composedWithGlobal:(MatrixTransform2x3 withOffset: 0@cy negated).
	transform _ transform composedWithGlobal:(MatrixTransform2x3 withScale: 1.0@-1.0).
	transform _ transform composedWithGlobal:(MatrixTransform2x3 withOffset: 0@cy).
	^transform! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTSampleStringMorph class
	instanceVariableNames: ''!

!TTSampleStringMorph class methodsFor: 'parts bin' stamp: 'tk 12/10/2001 17:34'!
descriptionForPartsBin
	^ self partName:	'TrueType banner'
		categories:		#('Text' ' Basic 2 ')
		documentation:	'A short text in a beautiful font.  Use the resize handle to change size.'! !

Object subclass: #TTFontDescription
	instanceVariableNames: 'glyphTable glyphs kernPairs copyright familyName fullName subfamilyName uniqueName versionName postscriptName trademark bounds unitsPerEm ascender descender lineGap '
	classVariableNames: 'Default Descriptions '
	poolDictionaries: ''
	category: 'Balloon-TrueType Fonts'!
!TTFontDescription commentStamp: '<historical>' prior: 0!
Holds a TrueType font in memory.  Is used by TTSampleStringMorph as its font.  

Class owns a default example.  !


!TTFontDescription methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:08'!
at: aCharOrInteger
	^glyphTable at: aCharOrInteger asInteger+1! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:08'!
at: index put: value
	^self shouldNotImplement! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
name

	^ self familyName copyWithout: Character space.
! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
size

	^ glyphs size.
! !


!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
ascender
	^ascender! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
bounds
	^bounds! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
descender
	^descender! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
lineGap
	^lineGap! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:49'!
unitsPerEm
	^unitsPerEm! !


!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
copyright
	^copyright! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
familyName
	^familyName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
fullName
	^fullName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
postscriptName
	^postscriptName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
subfamilyName
	^subfamilyName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
trademark
	^trademark! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
uniqueName
	^uniqueName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:49'!
versionName
	^versionName! !


!TTFontDescription methodsFor: 'converting' stamp: 'ar 11/14/1998 23:47'!
asMorph
	^TTSampleFontMorph font: self! !

!TTFontDescription methodsFor: 'converting' stamp: 'sma 5/5/2000 13:46'!
asStrikeFontScale: scale
	"Generate a StrikeFont (actually a FormSetFont) for this TTF font at a given scale."

	| forms |
	forms _ (0 to: 255) collect:
		[:i |
		(self at: i)
			asFormWithScale: scale
			ascender: ascender
			descender: descender].
	^ FormSetFont new
		fromFormArray: forms
		asciiStart: 0
		ascent: (ascender * scale) rounded! !


!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/14/1998 20:20'!
flipAroundY
	bounds _ (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	glyphs do:[:glyph| glyph flipAroundY]! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:27'!
setAscender: asc descender: desc lineGap: lgap
	ascender _ asc.
	descender _ desc.
	lineGap _ lgap! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:28'!
setBounds: aRect unitsPerEm: aNumber
	bounds _ aRect.
	unitsPerEm _ aNumber.! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:27'!
setGlyphs: glyphArray mapping: mappingTable
	glyphs _ glyphArray.
	glyphTable _ mappingTable.! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:48'!
setKernPairs: array
	kernPairs _ array! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:46'!
setStrings: anArray
	copyright _ anArray at: 1.
	familyName _ anArray at: 2.
	subfamilyName _ anArray at: 3.
	uniqueName _ anArray at: 4.
	fullName _ anArray at: 5.
	versionName _ anArray at: 6.
	postscriptName _ anArray at: 7.
	trademark _ anArray at: 8.
! !


!TTFontDescription methodsFor: 'copying' stamp: 'yo 6/23/2003 18:23'!
deepCopy

	"Since it shouldn't be copied for transmitting or any reason, it returns self."
	^ self.
! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 11/30/2002 22:38'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to a known Font in the other system instead.  "

	"A path to me"
	(TextConstants at: #forceFontWriting ifAbsent: [false]) ifTrue: [^ self].
		"special case for saving the default fonts on the disk.  See collectionFromFileNamed:"

	dp _ DiskProxy global: #TTFontDescription selector: #descriptionNamed:
			args: {self name}.
	refStrm replace: self with: dp.
	^ dp.
! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 11/30/2002 22:38'!
veryDeepCopyWith: deepCopier
	"Return self.  I am shared.  Do not record me."
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTFontDescription class
	instanceVariableNames: ''!

!TTFontDescription class methodsFor: 'instance creations' stamp: 'tb 6/24/2003 17:10'!
addFromTTFile: fileName
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"
	^self addFromTTStream: (FileStream readOnlyFileNamed: fileName).
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'tb 6/24/2003 17:08'!
addFromTTStream: readStream
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"

	| tt old |
	tt _ TTFontReader readFrom: readStream.
	old _ Descriptions detect: [:f | f name = tt name and: [f subfamilyName = tt subfamilyName]] ifNone: [nil].
	old ifNotNil: [Descriptions remove: old].
	Descriptions add: tt.
	^ tt.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
clearDefault
"
	self clearDefault
"

	Default _ nil.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
clearDescriptions
"
	self clearDescriptions
"

	Descriptions _ Set new.
	Default ifNotNil: [Descriptions add: Default].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'tk 12/10/2001 17:12'!
default
	^ Default! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
descriptionNamed: descriptionName

	^ Descriptions detect: [:f | f name = descriptionName] ifNone: [Default].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
initialize
"
	self initialize
"

	self clearDescriptions.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:18'!
removeDescriptionNamed: descriptionName

	| tts |
	Descriptions ifNil: [^ self].
	tts _ Descriptions select: [:f | f name = descriptionName].
	tts do: [:f | Descriptions remove: f].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:55'!
removeDescriptionNamed: descriptionName subfamilyName: subfamilyName

	| tts |
	Descriptions ifNil: [^ self].
	tts _ Descriptions select: [:f | f name = descriptionName and: [f subfamilyName = subfamilyName]].
	tts do: [:f | Descriptions remove: f].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:20'!
setDefault
"
	self setDefault
"

	Default _ TTFontReader readFrom: (FileStream readOnlyFileNamed: 'C:\WINDOWS\Fonts\comic.ttf').
! !



Object subclass: #TTGlyph
	instanceVariableNames: 'bounds contours advanceWidth leftSideBearing rightSideBearing glyphIndex '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Fonts'!
!TTGlyph commentStamp: '<historical>' prior: 0!
This class represents a glyph of a TrueType font.

Instance variables:
	bounds			<Rectangle>	The receiver's bounds
	contours		<Array of: PointArray> The compressed contours in the receiver
	advanceWidth	<Integer>	advance width of the glyph
	leftSideBearing	<Integer>	left side bearing
	rightSideBearing <Integer>	right side bearing
	glyphIndex 		<Integer>	the original index of the glyph (used for kerning)!


!TTGlyph methodsFor: 'initialize-release' stamp: 'ar 11/1/1998 22:25'!
initialize

	bounds _ 0@0 corner: 0@0.
	contours _ #().
	advanceWidth _ 0.
	leftSideBearing _ 0.
	rightSideBearing _ 0.! !


!TTGlyph methodsFor: 'accessing'!
advanceWidth
	^advanceWidth! !

!TTGlyph methodsFor: 'accessing'!
advanceWidth: aNumber
	advanceWidth _ aNumber.! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:25'!
bounds
	^bounds! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:25'!
bounds: aRectangle
	bounds _ aRectangle! !

!TTGlyph methodsFor: 'accessing'!
contours
	^contours! !

!TTGlyph methodsFor: 'accessing'!
contours: aCollection
	contours _ aCollection asArray.! !

!TTGlyph methodsFor: 'accessing'!
glyphIndex
	^glyphIndex! !

!TTGlyph methodsFor: 'accessing'!
glyphIndex: anInteger
	glyphIndex _ anInteger! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:26'!
glyphsAndTransformationsDo: aBlock
	aBlock value: self value: MatrixTransform2x3 identity! !

!TTGlyph methodsFor: 'accessing'!
leftSideBearing
	^leftSideBearing! !

!TTGlyph methodsFor: 'accessing'!
leftSideBearing: aNumber
	leftSideBearing _ aNumber.! !

!TTGlyph methodsFor: 'accessing'!
rightSideBearing
	^rightSideBearing! !

!TTGlyph methodsFor: 'accessing'!
rightSideBearing: aNumber
	rightSideBearing _ aNumber.! !


!TTGlyph methodsFor: 'testing'!
isComposite
	^false! !


!TTGlyph methodsFor: 'printing' stamp: 'tk 9/13/1999 09:54'!
printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPut:$(;
		print: (contours ifNil: [0] ifNotNil: [contours size]);
		nextPut:$).! !


!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:18'!
buildContours
	"Build the contours in the receiver glyph.
	The contour is constructed by converting the points
	form each contour into an absolute value and then
	compressing the contours into PointArrays."
	| tx ty points |
	tx _ ty _ 0.
	contours _ contours collect:[:contour|
		points _ contour points.
		points do:[:pt|
			pt x: (tx _ tx + pt x).
			pt y: (ty _ ty + pt y)].
		contour asCompressedPoints].! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:42'!
initializeContours: numContours with: endPoints
	"Initialize the contours for creation of the glyph."
	| startPt pts endPt |
	contours _ Array new: numContours.
	startPt _ -1.
	1 to: numContours do:[:i|
		endPt _ endPoints at: i.
		pts _ Array new: endPt - startPt.
		1 to: pts size do:[:j| pts at: j put: TTPoint new].
		contours at: i put: (TTContourConstruction on: pts).
		startPt _ endPt].! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:27'!
updateRightSideBearing
	"Update the right side bearing value"
	"@@: Is the following really correct?!!?!!"
	rightSideBearing _ advanceWidth - leftSideBearing - bounds corner x + bounds origin x! !


!TTGlyph methodsFor: 'private' stamp: 'ar 5/25/2000 18:01'!
display
	| canvas |
	canvas _ Display getCanvas.
	self contours do:[:ptArray|
		1 to: ptArray size by: 3 do:[:i|
			canvas line: (ptArray at: i) // 10
					to: (ptArray at: i+2) // 10
					width: 1 color: Color black.
		].
	].! !

!TTGlyph methodsFor: 'private' stamp: 'ar 11/14/1998 20:22'!
flipAroundY
	bounds _ (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	contours _ self contours collect:[:contour| contour collect:[:pt| pt x @ pt y negated]].! !


!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:29'!
asFormWithScale: scale ascender: ascender descender: descender
	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: Color black
		bgColor: Color white
		depth: 8
		replaceColor: true.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:27'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth

	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: fgColor
		bgColor: bgColor
		depth: depth
		replaceColor: false.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:26'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag

	| form canvas |
	form _ Form extent: (advanceWidth @ (ascender - descender) * scale) rounded depth: depth.
	form fillColor: bgColor.
	canvas _ BalloonCanvas on: form.
	canvas aaLevel: 4.
	canvas transformBy: (MatrixTransform2x3 withScale: scale asPoint * (1 @ -1)).
	canvas transformBy: (MatrixTransform2x3 withOffset: 0 @ ascender negated).
	canvas
		drawGeneralBezierShape: self contours
		color: fgColor 
		borderWidth: 0 
		borderColor: fgColor.
	replaceColorFlag ifTrue: [
		form replaceColor: bgColor withColor: Color transparent.
	].
	^ form! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTGlyph class
	instanceVariableNames: ''!

!TTGlyph class methodsFor: 'instance creation'!
new
	^super new initialize! !


TTGlyph subclass: #TTCompositeGlyph
	instanceVariableNames: 'glyphs '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Fonts'!

!TTCompositeGlyph methodsFor: 'initialize' stamp: 'ar 11/2/1998 01:20'!
initialize
	glyphs _ #().! !


!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:21'!
addGlyph: aGlyph transformation: aMatrix
	glyphs _ glyphs copyWith: (aMatrix -> aGlyph)! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:43'!
contours
	^contours ifNil:[contours _ self computeContours]! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:20'!
glyphs

	^glyphs collect:[:assoc| assoc value].! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:28'!
glyphsAndTransformationsDo: aBlock
	glyphs do:[:assoc|
		aBlock value: assoc value value: assoc key.
	].! !


!TTCompositeGlyph methodsFor: 'testing'!
isComposite
	^true! !


!TTCompositeGlyph methodsFor: 'private' stamp: 'ar 11/2/1998 01:43'!
computeContours
	| out |
	out _ WriteStream on: (Array new: glyphs size * 4).
	self glyphsAndTransformationsDo:[:glyph :transform|
		glyph contours do:[:ptArray|
			out nextPut: (transform localPointsToGlobal: ptArray).
		].
	].
	^out contents! !

!TTCompositeGlyph methodsFor: 'private' stamp: 'ar 11/14/1998 20:27'!
flipAroundY
	bounds _ (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	contours _ nil.! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTCompositeGlyph class
	instanceVariableNames: ''!

!TTCompositeGlyph class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 01:20'!
new
	^super new initialize! !


Object subclass: #TTKernPair
	instanceVariableNames: 'left right value mask '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Balloon-TrueType Fonts'!

!TTKernPair methodsFor: 'accessing'!
left
	^left! !

!TTKernPair methodsFor: 'accessing'!
left: aNumber

	left _ aNumber! !

!TTKernPair methodsFor: 'accessing' stamp: 'ar 11/1/1998 20:08'!
mask
	^mask ifNil:[mask _ self class maskFor: left with: right]! !

!TTKernPair methodsFor: 'accessing'!
right
	^right! !

!TTKernPair methodsFor: 'accessing'!
right: aNumber

	right _ aNumber! !

!TTKernPair methodsFor: 'accessing'!
value
	^value! !

!TTKernPair methodsFor: 'accessing'!
value: aNumber

	value _ aNumber! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

TTKernPair class
	instanceVariableNames: ''!

!TTKernPair class methodsFor: 'accessing'!
maskFor: left with: right
	^(left bitShift: 12) + right! !

TTFontDescription initialize!

'From Squeak3.5 of ''11 April 2003'' [latest update: #5180] on 23 June 2003 at 8:52 pm'!
"Change Set:		TrueTypeTextStyle6
Date:			23 June 2003
Author:			Yoshiki Ohshima

This change set will allow you to use TrueType based outline glyphs as a TextStyle.  Once you load this change set, you will see 'ComicSansMS' style in the 'Change font' halo menu of the TextMorphs.  You can add new styles by calling TTCFont>>newTextStyleFromTTFile: or, on 3.4, by selecting 'install ttf style' from the FileList.  You definitely will want to try symbol.ttf:-)

  On Mac, the fonts are not stored as separated files.  We should wait for the Mac version of NativeFont plugin.

  Thanks to Phil Hargett, the service mechanism for 3.4 FileList is added.  Also, following his suggestions, now you can use TTCFont in MVC as well.

  This change set was first written for the multilingualized Squeak and .TTC (TrueType Collection) file.  The name of the class and the category name are reminiscent of it.

  In the versions prior 4, there were two similar classes, but now those are merged into one.

  Also, in version 4, bold, italic, and bold italic faces can be handled properly.  Install the same family fonts and choose them from the Emphasis and Alignment halo.

  In version 5, you can now use truetype in Nebraska.  You have to install the fonts on both side before you launch Nebraska.  You can also add new font size for a style.

  In version 6, the confused pointSize <> pixelSize problem is sorted out.  However, it could add a compatibility problem."!

CharacterScanner subclass: #CanvasCharacterScanner
	instanceVariableNames: 'canvas fillBlt foregroundColor runX lineY defaultTextColor '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!
AbstractFont subclass: #TTCFont
	instanceVariableNames: 'ttcDescription pointSize foregroundColor cache derivatives '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Multilingual-Display'!

!TTCFont commentStamp: 'yo 6/23/2003 19:44' prior: 0!
I represent a font that uses TrueType derived glyph.  Upon a request for glyph for a character through a call to #formOf: (or #widthOf:), I first search corresponding glyph in the cache.  If there is not, it creates a 32bit depth form with the glyph.

  The cache is weakly held.  The entries are zapped at full GC.

Structure:
 ttcDescription	TTFontDescription -- The Squeak data structure for a TrueType font data file.
 pointSize		Number -- character size in point.  It simply assumes that the resolusion of display device is 96 dpi and an inch is 72 point.
 foregroundColor	Color -- So far, this font need to know the glyph color in cache.
 cache			WeakArray pointing to Array(256).
 derivatives		Array -- stores the fonts in the same family but different emphasis.
!

Object subclass: #TTFontDescription
	instanceVariableNames: 'glyphTable glyphs kernPairs copyright familyName fullName subfamilyName uniqueName versionName postscriptName trademark bounds unitsPerEm ascender descender lineGap '
	classVariableNames: 'Default Descriptions '
	poolDictionaries: ''
	category: 'Balloon-TrueType Fonts'!

!AbstractFont methodsFor: 'accessing' stamp: 'nk 5/26/2003 09:45'!
isRegular
	^false! !


!BitBlt methodsFor: 'private' stamp: 'yo 6/23/2003 18:07'!
installTTCFont: aTTCFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	"Set up the parameters.  Since the glyphs in a TTCFont is 32bit depth form, it tries to use rule=34 to get better AA result if possible."

	((aTTCFont depth = 32)) ifTrue: [
		destForm depth <= 8 ifTrue: [
			self colorMap: (self cachedFontColormapFrom: aTTCFont depth to: destForm depth).
			self combinationRule: Form paint.
		] ifFalse: [
			self colorMap: nil.
			self combinationRule: 34.
		].
		halftoneForm _ nil.
		sourceY _ 0.
		height _ aTTCFont height.
		^ self.
	].
! !



!CanvasCharacterScanner methodsFor: 'private' stamp: 'yo 6/23/2003 18:09'!
defaultTextColor: color
	"This defaultTextColor inst var is equivalent to paragraphColor of DisplayScanner."
	defaultTextColor _ color.
! !

!CanvasCharacterScanner methodsFor: 'private' stamp: 'yo 6/8/2003 21:29'!
setFont

	foregroundColor _ defaultTextColor.
	super setFont.
	destY _ lineY + line baseline - font ascent! !


!CanvasDecoder methodsFor: 'decoding' stamp: 'yo 3/21/2003 23:02'!
addTTCFontToCache: command
	| index font |
	index := self class decodeInteger: command second.
	font := self class decodeTTCFont: command third.

	index > fonts size ifTrue: [
		| newFonts |
		newFonts  := Array new: index.
		newFonts replaceFrom: 1 to: fonts size with: fonts.
		fonts := newFonts ].

	fonts at: index put: font.
! !

!CanvasDecoder methodsFor: 'decoding' stamp: 'yo 3/21/2003 23:00'!
processCommand: command  onForceDo: forceBlock
	| verb verbCode |
	command isEmpty ifTrue: [ ^self ].

	verb _ command first.
	verbCode := verb at: 1.

	verbCode = CanvasEncoder codeClip ifTrue: [ ^self setClip: command ].
	verbCode = CanvasEncoder codeTransform ifTrue: [ ^self setTransform: command ].
	verbCode = CanvasEncoder codeText ifTrue: [ ^self drawText: command ].
	verbCode = CanvasEncoder codeLine ifTrue: [ ^self drawLine: command ].
	verbCode = CanvasEncoder codeRect ifTrue: [ ^self drawRect: command ].
	verbCode = CanvasEncoder codeBalloonRect ifTrue: [ ^self drawBalloonRect: command ].
	verbCode = CanvasEncoder codeBalloonOval ifTrue: [ ^self drawBalloonOval: command ].
	verbCode = CanvasEncoder codeInfiniteFill ifTrue: [ ^self drawInfiniteFill: command ].

	verbCode = CanvasEncoder codeOval ifTrue: [ ^self drawOval: command ].
	verbCode = CanvasEncoder codeImage ifTrue: [ ^self drawImage: command ].
	verbCode = CanvasEncoder codeReleaseCache ifTrue: [ ^self releaseImage: command ].
	verbCode = CanvasEncoder codePoly ifTrue: [ ^self drawPoly: command ].
	verbCode = CanvasEncoder codeStencil ifTrue: [ ^self drawStencil: command ].
	verbCode = CanvasEncoder codeForce ifTrue: [ ^self forceToScreen: command withBlock: forceBlock ].
	verbCode = CanvasEncoder codeFont ifTrue: [ ^self addFontToCache: command ].
	verbCode = CanvasEncoder codeTTCFont ifTrue: [ ^self addTTCFontToCache: command ].
	verbCode = CanvasEncoder codeExtentDepth ifTrue: [ ^self extentDepth: command ].
	verbCode = CanvasEncoder codeShadowColor ifTrue: [ ^self shadowColor: command ].

self error: 'unknown command: ', command first.! !


!CanvasDecoder class methodsFor: 'decoding' stamp: 'yo 6/23/2003 20:12'!
decodeTTCFont: fontString

	"Decode a string that consists of <familyName> <pointSize> <emphasis> (e.g. 'ComicSansMS 12 0') into a proper instance."

	| first second |
	first _ fontString indexOf: $  startingAt: 1.
	second _ fontString indexOf: $  startingAt: first + 1.

	(first ~= 0 and: [second ~= 0]) ifTrue: [
		^ (TTCFont family: (fontString copyFrom: 1 to: (first - 1))
			size: (fontString copyFrom: first + 1 to: second - 1) asNumber)
				emphasized: (fontString copyFrom: second + 1 to: fontString size) asNumber.
	].

	^ TextStyle defaultFont.
! !


!CanvasEncoder methodsFor: 'fonts' stamp: 'yo 6/23/2003 20:04'!
sendFont: aFont atIndex: index
	"Transmits the given fint to the other side"

	| code |
	code _ CanvasEncoder codeFont.
	aFont class = TTCFont ifTrue: [code _ CanvasEncoder codeTTCFont].
	self sendCommand: {
		String with: code.
		self class encodeInteger: index.
		self class encodeFont: aFont }.
! !


!CanvasEncoder class methodsFor: 'codes' stamp: 'yo 3/21/2003 23:00'!
codeTTCFont

	^ $T.
! !

!CanvasEncoder class methodsFor: 'as yet unclassified' stamp: 'yo 3/21/2003 23:01'!
nameForCode: aStringOrChar

	| ch |
	ch _ (aStringOrChar isKindOf: String) ifTrue: [aStringOrChar first] ifFalse: [aStringOrChar].
	ch == self codeBalloonOval ifTrue: [^'balloon oval'].
	ch == self codeBalloonRect ifTrue: [^'balloon rectangle'].
	ch == self codeClip ifTrue: [^'clip'].
	ch == self codeExtentDepth ifTrue: [^'codeExtentDepth'].
	ch == self codeFont ifTrue: [^'codeFont'].
	ch == self codeTTCFont ifTrue: [^'codeTTCFont'].
	ch == self codeForce ifTrue: [^'codeForce'].
	ch == self codeImage ifTrue: [^'codeImage'].
	ch == self codeLine ifTrue: [^'codeLine'].
	ch == self codeOval ifTrue: [^'codeOval'].
	ch == self codePoly ifTrue: [^'codePoly'].
	ch == self codeRect ifTrue: [^'codeRect'].
	ch == self codeReleaseCache ifTrue: [^'codeReleaseCache'].
	ch == self codeStencil ifTrue: [^'codeStencil'].
	ch == self codeText ifTrue: [^'codeText'].
	ch == self codeTransform ifTrue: [^'codeTransform'].
	ch == self codeInfiniteFill ifTrue: [^'codeInfiniteFill'].
	ch == self codeShadowColor ifTrue: [^'shadowColor'].
	^'????'
! !


!DisplayText methodsFor: 'displaying' stamp: 'yo 6/23/2003 20:05'!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: ruleInteger fillColor: aForm
	"For TT font, rule 34 is used if possible."
	"Refer to the comment in 
	DisplayObject|displayOn:at:clippingBox:rule:mask:."

	| form1 rule |
	form1 _ self form.
	rule _ (ruleInteger = Form over and: [backColor isTransparent])
				ifTrue: [form1 depth = 32 ifTrue: [rule _ 34] ifFalse: [Form paint]]
				ifFalse: [ruleInteger].
	form1 depth = 32 ifTrue: [rule _ 34].
	form1
		displayOn: aDisplayMedium
		at: aDisplayPoint + offset
		clippingBox: clipRectangle
		rule: rule
		fillColor: aForm! !

!DisplayText methodsFor: 'private' stamp: 'yo 6/8/2003 21:47'!
composeForm
	"For the TT strings in MVC widget in Morphic world such as a progress bar, form is create by Morphic machinary."
	| canvas tmpText |
	Smalltalk isMorphic
		ifTrue:
			[tmpText _ TextMorph new contentsAsIs: text deepCopy.
			foreColor ifNotNil: [tmpText text addAttribute: (TextColor color: foreColor)].
			backColor ifNotNil: [tmpText backgroundColor: backColor].
			tmpText setTextStyle: textStyle.
			canvas _ FormCanvas on: (Form extent: tmpText extent depth: 32).
			tmpText drawOn: canvas.
			form _ canvas form.
		]
		ifFalse: [form _ self asParagraph asForm]! !


!GrafPort methodsFor: 'private' stamp: 'yo 6/23/2003 20:34'!
installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	super installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor.
	alpha _ foregroundColor privateAlpha.
	"dynamically switch between blend modes to support translucent text"
	"To handle the transition from TTCFont to StrikeFont, rule 34 must be taken into account."
	alpha = 255 ifTrue:[
		combinationRule = 30 ifTrue: [combinationRule _ Form over].
		combinationRule = 31 ifTrue: [combinationRule _ Form paint].
		combinationRule = 34 ifTrue: [combinationRule _ Form paint].
	] ifFalse:[
		combinationRule = Form over ifTrue: [combinationRule _ 30].
		combinationRule = Form paint ifTrue: [combinationRule _ 31].
		combinationRule = 34 ifTrue: [combinationRule _ 31].
	]! !


!PopUpMenu methodsFor: 'private' stamp: 'yo 12/13/2002 15:47'!
computeForm
	"Compute and answer a Form to be displayed for this menu."

	| borderInset paraForm menuForm inside |
	borderInset _ 2@2.
	paraForm _ (DisplayText text: labelString asText textStyle: MenuStyle) form.
	menuForm _ Form extent: paraForm extent + (borderInset * 2) depth: paraForm depth.
	menuForm borderWidth: 2.
	paraForm displayOn: menuForm at: borderInset.
	lineArray == nil ifFalse:
		[lineArray do:
			[ :line |
			menuForm fillBlack: (4 @ ((line * font height) + borderInset y)
				extent: (menuForm width - 8 @ 1))]].

	frame _ Quadrangle new.
	frame region: menuForm boundingBox.
	frame borderWidth: 4.
	inside _ frame inside.
	marker _ inside topLeft extent: (inside width @ MenuStyle lineGrid).
	selection _ 1.

	^ form _ menuForm
! !


!RemoteCanvas methodsFor: 'as yet unclassified' stamp: 'yo 6/23/2003 18:09'!
paragraph: paragraph bounds: bounds color: c

	| scanner |
	scanner _ CanvasCharacterScanner new.
	scanner
		 canvas: self;
		text: paragraph text textStyle: paragraph textStyle;
		textColor: c; defaultTextColor: c.

	paragraph displayOn: self using: scanner at: bounds topLeft.
! !


!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 6/23/2003 19:05'!
fromUser: priorFont
	"Present a menu of available fonts, and if one is chosen, return it."
	| fontList fontMenu style active ptMenu label spec font |
	fontList _ StrikeFont familyNames remove: 'DefaultTextStyle' ifAbsent: []; asOrderedCollection.
	fontMenu _ MenuMorph new defaultTarget: self.
	fontList do: [:fontName |
		style _ TextStyle named: fontName.
		active _ priorFont familyName sameAs: fontName.
		ptMenu _ MenuMorph new defaultTarget: self.
		style pointSizes do: [:pt |
			(active and:[pt = priorFont pointSize]) 
				ifTrue:[label _ '<on>'] 
				ifFalse:[label _ '<off>'].
			label _ label, pt printString, ' pt'.
			ptMenu add: label 
				target: fontMenu
				selector: #modalSelection:
				argument: {fontName. pt}].
		style isTTCStyle ifTrue: [
			ptMenu add: 'new size'
				target: style selector: #addNewFontSizeDialog: argument: {fontName. fontMenu}.
		].
		active ifTrue:[label _ '<on>'] ifFalse:[label _ '<off>'].
		label _ label, fontName.
		fontMenu add: label subMenu: ptMenu].
	spec _ fontMenu invokeModal: false. "don't allow keyboard control"
	spec ifNil: [^ nil].
	style _ TextStyle named: spec first.
	style ifNil: [^ self].
	font _ style fonts detect: [:any | any pointSize = spec last] ifNone: [nil].
	^ font! !


!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:42'!
ascent

	^ ttcDescription ascender * self pixelSize // (ttcDescription ascender - ttcDescription descender).
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
baseKern

	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 20:13'!
descent
	"One is added to make sure the gap between lines is filled.  If we don't add, multi line selection in a text pane look ugly."
	^ ttcDescription descender * self pixelSize // (ttcDescription descender - ttcDescription ascender) + 1.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:39'!
descentKern

	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 3/22/2003 13:32'!
emphasis

	ttcDescription subfamilyName = 'Regular' ifTrue: [^ 0].
	ttcDescription subfamilyName = 'Bold' ifTrue: [^ 1].
	ttcDescription subfamilyName = 'Italic' ifTrue: [^ 2].
	ttcDescription subfamilyName = 'Bold Italic' ifTrue: [^ 3].
	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 5/26/2003 20:13'!
emphasis: code

	code > 3 ifTrue: [^ self].
	code = 0 ifTrue: [^ self].
	derivatives isNil ifTrue: [^ self].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 12/13/2002 12:46'!
emphasized: code

	code > 3 ifTrue: [^ self].
	code = 0 ifTrue: [^ self].
	derivatives isNil ifTrue: [^ self].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 12/10/2002 17:08'!
familyName

	^ ttcDescription name.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:39'!
familySizeFace

	^ Array
		with: self familyName
		with: self height
		with: 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:39'!
height

	^ self pixelSize.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 12/12/2002 21:04'!
isRegular

	^ ttcDescription subfamilyName = 'Regular'.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
maxAscii

	^ ttcDescription size.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
minAscii

	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
name

	^ ttcDescription name.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:41'!
pixelSize

	^ pointSize * 96 / 72.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:39'!
pointSize

	^ pointSize.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:41'!
pointSize: aNumber

	pointSize _ aNumber.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 12/2/2002 14:09'!
textStyle
	^ Utilities actualTextStyles detect:
		[:aStyle | aStyle fontArray includes: self] ifNone: [nil]! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:35'!
depth

	^ 32.
! !

!TTCFont methodsFor: 'public' stamp: 'yo 11/16/2002 00:59'!
foregroundColor

	^ foregroundColor.
! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:36'!
size

	^ ttcDescription size.
! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:36'!
widthOf: aCharacter

	"This method cannot use #formOf: because formOf: discriminates the color and causes unnecessary bitmap creation."

	| f assoc |
	assoc _ self cache at: (aCharacter asciiValue + 1).
	assoc ifNotNil: [
		^ assoc value width
	].

	f _ self computeForm: aCharacter.
	self at: aCharacter put: f.
	^ f width.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/10/2002 17:47'!
cache

	(cache at: 1) ifNil: [self flushCache].
	^ cache at: 1.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/13/2002 12:53'!
derivativeFont: aTTCFont

	| index |
	aTTCFont ifNil: [derivatives _ nil. ^ self].
	derivatives ifNil: [derivatives _ Array new: 4].
	index _ #('Bold' 'Italic' 'Bold Italic') indexOf: (aTTCFont ttcDescription subfamilyName).
	index = 0 ifTrue: [
		^ self inform: 'unknown sub family name.  This font will be skipped'.
	].

	derivatives at: index put: aTTCFont.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/13/2002 12:13'!
derivativeFonts

	| result |
	derivatives ifNil: [^ #()].
	result _ OrderedCollection new.
	derivatives do: [:f | result add: f].
	^ result asArray.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/10/2002 16:35'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta 

	| destPoint ascii form |
	destPoint _ aPoint.
	startIndex to: stopIndex do: [:charIndex | 
		ascii _ (aString at: charIndex) asciiValue bitAnd: 16rFFFFFF.
		((ascii between: self minAscii and: self maxAscii) not) ifTrue: [
			ascii _ self maxAscii].
		form _ self formOf: (aString at: charIndex).
		aBitBlt sourceForm: form.
		aBitBlt destX: destPoint x.
		aBitBlt destY: destPoint y.
		aBitBlt sourceOrigin: 0 @ 0.
		aBitBlt width: form width.
		aBitBlt height: form height.
		aBitBlt copyBits.
		destPoint _ destPoint + (form width + kernDelta @ 0).
	].
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/10/2002 16:35'!
flushCache

	cache at: 1 put: (Array new: 256).
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 11/16/2002 01:00'!
initialize

	foregroundColor _ Color black.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/10/2002 18:15'!
installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor

	foregroundColor _ fgColor.

	aDisplayContext installTTCFont: self foregroundColor: foregroundColor backgroundColor: bgColor
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/10/2002 16:36'!
recreateCache

	cache _ WeakArray new: 1.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 11/16/2002 01:01'!
ttcDescription

	^ ttcDescription.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 12/10/2002 17:27'!
ttcDescription: aTTCDescription

	ttcDescription _ aTTCDescription.
	self recreateCache.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 12/10/2002 16:34'!
at: char put: form

	self cache at: (char asciiValue + 1) put: (foregroundColor -> form).
! !

!TTCFont methodsFor: 'private' stamp: 'yo 6/23/2003 18:39'!
computeForm: char

	| ttGlyph scale |
	scale _ self pixelSize asFloat / (ttcDescription ascender - ttcDescription descender).
	ttGlyph _ ttcDescription at: char.
	^ ttGlyph asFormWithScale: scale ascender: ttcDescription ascender descender: ttcDescription descender fgColor: foregroundColor bgColor: Color transparent depth: self depth.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 12/10/2002 16:35'!
formOf: char

	| f assoc |
	assoc _ self cache at: (char asciiValue + 1).
	assoc ifNotNil: [
		(assoc key = foregroundColor) ifTrue: [
			^ assoc value.
		].
	].

	f _ self computeForm: char.
	self at: char put: f.
	^ f.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:35'!
copy

	^ self.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:35'!
deepCopy

	^ self.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:36'!
objectForDataStream: refStrm

	self flushCache.
	^ self.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:36'!
veryDeepCopyWith: deepCopier

	self flushCache.
	^ self.
! !

!TTCFont methodsFor: 'file in/out' stamp: 'yo 6/23/2003 18:44'!
encodedForRemoteCanvas

	^ self familyName, ' ', self pointSize printString, ' ', self emphasis printString.
! !


!TTCFont class methodsFor: 'instance creation' stamp: 'yo 6/23/2003 18:44'!
family: f size: s

	^ self allInstances detect: [:a | a familyName = f and: [a pointSize = s]] ifNone: [nil].
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 12/13/2002 12:20'!
getExistings: fontArray

	| result em |
	result _ OrderedCollection new.
	result add: fontArray.
	1 to: 3 do: [:i |
		em _ (fontArray collect: [:f | f emphasized: i]).
		(em at: 1) ~= (fontArray at: 1) ifTrue: [
			result add: em.
		].
	].
	^ result asArray.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 6/23/2003 20:17'!
newTextStyleFromTT: description
	"Create a new TextStyle from specified TTFontDescription instance."
	| array f |

	array _ self pointSizes collect: [:pt |
		f _ TTCFont new initialize.
		f ttcDescription: description.
		f pointSize: pt.
	].

	^ self reorganizeForNewFontArray: array.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 6/23/2003 20:17'!
newTextStyleFromTTFile: fileName
	"Create a new TextStyle from specified file name.  On certain versions of Windows, you can evaluate following to get Arial font into the image.  On other platforms, wait and see someone implements the support code for FontPlugin then we can start relying on the generic font lookup mechanism.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\ARIAL.TTF'
	"

	| description |
	description _ TTFontDescription addFromTTFile: fileName.
	^ self newTextStyleFromTT: description.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 12/13/2002 14:22'!
reorganizeForNewFontArray: array

	| styleName style existings regular |
	styleName _ (array at: 1) name asSymbol.
	(TextConstants includesKey: styleName) ifFalse: [
		TextConstants at: styleName put: (TextStyle fontArray: array).
		^ TextConstants at: styleName.
	].
 
	style _ TextConstants at: styleName.
	existings _ (self getExistings: style fontArray), (Array with: array).
	regular _ existings detect: [:e | (e at: 1) isRegular] ifNone: [existings at: 1].

	existings do: [:e |
		(e at: 1) isRegular ifFalse: [
			regular with: e do: [:r :f |
				r derivativeFont: f.
			].
		].
	].

	style newFontArray: regular.
	self recreateCache.	
	^ style.
! !

!TTCFont class methodsFor: 'class initialization' stamp: 'yo 6/23/2003 18:45'!
initialize
"
	self initialize
"

	| tt |
	tt _ TTFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: tt].

	(FileList respondsTo: #registerFileReader:) ifTrue: [
		FileList registerFileReader: self
	].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 19:46'!
isCacheAllNil
"
	self cacheAllNil
"
	self allInstances do: [:inst |
		inst cache do: [:e |
			e ifNotNil: [^ false].
		].
	].

	^ true.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 20:18'!
pointSizes

	"The default sizes that are created when a TextStyle is created.  You can add new sizes by the new-size feature."
	^ #(9 12 15 24 36).
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 12/10/2002 17:38'!
recreateCache
"
	self recreateCache.
"
	self allInstances do: [:inst | inst recreateCache].
	Smalltalk garbageCollect.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 12/13/2002 14:16'!
removeAllDerivatives
"
	self removeAllDerivatives
"

	self allInstances do: [:s |
		s textStyle ifNotNil: [
			s textStyle fontArray do: [:f |
				f derivativeFont: nil.
			].
		].
	].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 11/30/2002 22:37'!
removeStyleName: aString

	TextConstants removeKey: aString asSymbol ifAbsent: [].
	TTFontDescription removeDescriptionNamed: aString asString.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 19:50'!
version

	^ '6.0'.
! !

!TTCFont class methodsFor: 'file list services'!
fileReaderServicesForFile: fullName suffix: suffix 
	(suffix = 'ttf')
		ifTrue: [^self services].
	^ OrderedCollection new.! !

!TTCFont class methodsFor: 'file list services'!
serviceInstallTrueTypeFontStyle
	"Return a service to install a true type font as a text style"

	^ SimpleServiceEntry
		provider: self
		label: 'install ttf style'
		selector: #newTextStyleFromTTFile: 
		description: 'install a true type font as a text style'
		buttonLabel: 'install ttf'! !

!TTCFont class methodsFor: 'file list services'!
services
	"Return a set of services for use in FileList"

	^ Array with: self serviceInstallTrueTypeFontStyle! !


!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
name

	^ self familyName copyWithout: Character space.
! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
size

	^ glyphs size.
! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 6/23/2003 18:23'!
deepCopy

	"Since it shouldn't be copied for transmitting or any reason, it returns self."
	^ self.
! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 11/30/2002 22:38'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to a known Font in the other system instead.  "

	"A path to me"
	(TextConstants at: #forceFontWriting ifAbsent: [false]) ifTrue: [^ self].
		"special case for saving the default fonts on the disk.  See collectionFromFileNamed:"

	dp _ DiskProxy global: #TTFontDescription selector: #descriptionNamed:
			args: {self name}.
	refStrm replace: self with: dp.
	^ dp.
! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 11/30/2002 22:38'!
veryDeepCopyWith: deepCopier
	"Return self.  I am shared.  Do not record me."
! !


!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:16'!
addFromTTFile: fileName
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"

	| tt old |
	tt _ TTFontReader readFrom: (FileStream readOnlyFileNamed: fileName).
	old _ Descriptions detect: [:f | f name = tt name and: [f subfamilyName = tt subfamilyName]] ifNone: [nil].
	old ifNotNil: [Descriptions remove: old].
	Descriptions add: tt.
	^ tt.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
clearDefault
"
	self clearDefault
"

	Default _ nil.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
clearDescriptions
"
	self clearDescriptions
"

	Descriptions _ Set new.
	Default ifNotNil: [Descriptions add: Default].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
descriptionNamed: descriptionName

	^ Descriptions detect: [:f | f name = descriptionName] ifNone: [Default].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 11/30/2002 22:22'!
initialize
"
	self initialize
"

	self clearDescriptions.
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:18'!
removeDescriptionNamed: descriptionName

	| tts |
	Descriptions ifNil: [^ self].
	tts _ Descriptions select: [:f | f name = descriptionName].
	tts do: [:f | Descriptions remove: f].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:55'!
removeDescriptionNamed: descriptionName subfamilyName: subfamilyName

	| tts |
	Descriptions ifNil: [^ self].
	tts _ Descriptions select: [:f | f name = descriptionName and: [f subfamilyName = subfamilyName]].
	tts do: [:f | Descriptions remove: f].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/13/2002 13:20'!
setDefault
"
	self setDefault
"

	Default _ TTFontReader readFrom: (FileStream readOnlyFileNamed: 'C:\WINDOWS\Fonts\comic.ttf').
! !


!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:29'!
asFormWithScale: scale ascender: ascender descender: descender
	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: Color black
		bgColor: Color white
		depth: 8
		replaceColor: true.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:27'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth

	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: fgColor
		bgColor: bgColor
		depth: depth
		replaceColor: false.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:26'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag

	| form canvas |
	form _ Form extent: (advanceWidth @ (ascender - descender) * scale) rounded depth: depth.
	form fillColor: bgColor.
	canvas _ BalloonCanvas on: form.
	canvas aaLevel: 4.
	canvas transformBy: (MatrixTransform2x3 withScale: scale asPoint * (1 @ -1)).
	canvas transformBy: (MatrixTransform2x3 withOffset: 0 @ ascender negated).
	canvas
		drawGeneralBezierShape: self contours
		color: fgColor 
		borderWidth: 0 
		borderColor: fgColor.
	replaceColorFlag ifTrue: [
		form replaceColor: bgColor withColor: Color transparent.
	].
	^ form! !


!TextStyle methodsFor: 'accessing' stamp: 'yo 6/23/2003 19:02'!
isTTCStyle

	^ fontArray first isKindOf: TTCFont.
! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'yo 6/23/2003 19:57'!
addNewFontSize: pointSize
	"Add a font in specified size to the array of fonts."
	| f d newArray t |
	fontArray first emphasis ~= 0 ifTrue: [
		t _ TextConstants at: self fontArray first familyName asSymbol.
		t fonts first emphasis = 0 ifTrue: [
			^ t addNewFontSize: pointSize.
		].
	].

	pointSize <= 0 ifTrue: [^ nil].
	fontArray do: [:s |
		s pointSize = pointSize ifTrue: [^ s].
	].

	f _ TTCFont new initialize.
	f ttcDescription: fontArray first ttcDescription.
	f pointSize: pointSize.
	fontArray first derivativeFonts do: [:proto |
		proto ifNotNil: [
			d _ TTCFont new initialize.
			d ttcDescription: proto ttcDescription.
			d pointSize: f pointSize.
			f derivativeFont: d.
		].
	].
	newArray _ ((fontArray copyWith: f) asSortedCollection: [:a :b | a pointSize <= b pointSize]) asArray.
	self newFontArray: newArray.
	^ self fontOfPointSize: pointSize
! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'yo 6/23/2003 20:04'!
addNewFontSizeDialog: args
	"This is called from a modal menu and call back the menu with entered argument."
	| f n r |
	f _ FillInTheBlank request: 'Enter the point size' initialAnswer: '12'.
	n _ f asNumber.
	r _ self addNewFontSize: n.
	r ifNotNil: [
		args second ifNotNil: [args second modalSelection: {args first. n}].
	].
! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'yo 6/23/2003 19:58'!
discardOtherSizes
	"This method trys to discard the fonts in non-standard size.  If the size is still in use, there will be a problem."
	| newArray |
	self isTTCStyle ifFalse: [^ self].
	newArray _ fontArray select: [:s | TTCFont pointSizes includes: s pointSize].
	self newFontArray: newArray.

"(TextConstants at: #ComicSansMS) discardOtherSizes"! !


!Utilities class methodsFor: 'text styles and fonts' stamp: 'yo 6/23/2003 20:07'!
chooseTTCFontSize: args
	"Adds the new-size selection to the font size menu."
	| f n r |
	f _ FillInTheBlank request: 'New Point Size' initialAnswer: '0'.
	n _ f asNumber.
	r _ (TextConstants at: args first) addNewFontSize: n.
	r ifNotNil: [
		args second ifNotNil: [args second perform: args third with: r].
	].
! !

!Utilities class methodsFor: 'text styles and fonts' stamp: 'yo 6/23/2003 19:04'!
fontMenuForStyle: styleName target: target selector: selector
 	| aMenu |
	aMenu _ MenuMorph entitled: styleName.
	TextConstants at: styleName ifPresent: [:s |
		s isTTCStyle ifTrue: [
			aMenu add: 'New Size' target: self selector: #chooseTTCFontSize: argument: {styleName. target. selector}.
		].
	].
 	(Utilities pointSizesFor: styleName) do:
		[:aWidth |
			aMenu 
				add: (aWidth asString, ' Point') 
				target: target 
				selector: selector
				argument: ((TextStyle named: styleName) fontOfPointSize: aWidth).
                aMenu lastItem font: ((TextStyle named: styleName) fontOfPointSize: aWidth)].
        ^ aMenu! !

TTGlyph removeSelector: #asFormWithScale:ascender:descender:fgColor:bgColor:!
TTFontDescription initialize!

!TTFontDescription class reorganize!
('instance creations' addFromTTFile: clearDefault clearDescriptions default descriptionNamed: initialize removeDescriptionNamed: removeDescriptionNamed:subfamilyName: setDefault)
('upgrade patch')
!

TTCFont initialize!

!TTCFont class reorganize!
('instance creation' family:size: getExistings: newTextStyleFromTT: newTextStyleFromTTFile: reorganizeForNewFontArray:)
('class initialization' initialize)
('other' isCacheAllNil pointSizes recreateCache removeAllDerivatives removeStyleName: version)
('file list services' fileReaderServicesForFile:suffix: serviceInstallTrueTypeFontStyle services)
!


!TTCFont reorganize!
('accessing' ascent baseKern descent descentKern emphasis emphasis: emphasized: familyName familySizeFace height isRegular maxAscii minAscii name pixelSize pointSize pointSize: textStyle)
('public' depth foregroundColor size widthOf:)
('friend' cache derivativeFont: derivativeFonts displayString:on:from:to:at:kern: flushCache initialize installOn:foregroundColor:backgroundColor: recreateCache ttcDescription ttcDescription:)
('private' at:put: computeForm: formOf:)
('copying' copy deepCopy objectForDataStream: veryDeepCopyWith:)
('file in/out' encodedForRemoteCanvas)
!

!FileList methodsFor: 'file list menu' stamp: 'yo 2/19/2004 22:52'!
itemsForFileEnding: suffix 
	| labels lines selectors |
	labels _ OrderedCollection new.
	lines _ OrderedCollection new.
	selectors _ OrderedCollection new.
	suffix = 'bmp' | (suffix = 'gif') | (suffix = 'jpg') | (suffix = 'form') | (suffix = '*') | (suffix = 'png')
		ifTrue: [labels addAll: #('open image in a window' 'read image into ImageImports' 'open image as background' ).
			selectors addAll: #(#openImageInWindow #importImage #openAsBackground )].
	suffix = 'morph' | (suffix = 'morphs') | (suffix = 'sp') | (suffix = '*')
		ifTrue: [labels add: 'load as morph'.
			selectors add: #openMorphFromFile.
			labels add: 'load as project'.
			selectors add: #openProjectFromFile].
	suffix = 'mdl'
		ifTrue: [labels add: 'load into Wonderland'.
			selectors add: #openModelintoAlice].
	suffix = 'extseg' | (suffix = 'project') | (suffix = 'pr')
		ifTrue: [labels add: 'load as project'.
			selectors add: #openProjectFromFile].
	suffix = 'bo' | (suffix = '*')
		ifTrue: [labels add: 'load as book'.
			selectors add: #openBookFromFile].
	suffix = 'mid' | (suffix = '*')
		ifTrue: [labels add: 'play midi file'.
			selectors add: #playMidiFile].
	suffix = 'movie' | (suffix = '*')
		ifTrue: [labels add: 'open as movie'.
			selectors add: #openAsMovie].
	suffix = 'changes' | (suffix = '*')
		ifTrue: [labels add: 'browse recent changes'.
			selectors add: #browseRecentChanges].
	suffix = 'st' | (suffix = 'cs') | (suffix = '*')
		ifTrue: [suffix = '*'
				ifTrue: [lines add: labels size].
			labels addAll: #('fileIn' 'file into new change set' 'browse changes' 'browse code' 'remove line feeds' 'broadcast as update' ).
			lines add: labels size - 1.
			selectors addAll: #(#fileInSelection #fileIntoNewChangeSet #browseChanges #browseFile #removeLinefeeds #putUpdate )].
	suffix = 'swf' | (suffix = '*')
		ifTrue: [labels add: 'open as Flash'.
			selectors add: #openAsFlash].
	suffix = 'ttf' | (suffix = '*')
		ifTrue: [labels add: 'install true type font'.
			selectors add: #installTTFStyle].
	suffix = 'gz' | (suffix = '*')
		ifTrue: [labels addAll: #('view decompressed' 'decompress to file' ).
			selectors addAll: #(#viewGZipContents #saveGZipContents )].
	suffix = '3ds' | (suffix = '*')
		ifTrue: [labels add: 'Open 3DS file'.
			selectors add: #open3DSFile].
	suffix = 'tape' | (suffix = '*')
		ifTrue: [labels add: 'open for playback'.
			selectors add: #openTapeFromFile].
	suffix = 'wrl' | (suffix = '*')
		ifTrue: [labels add: 'open in Wonderland'.
			selectors add: #openVRMLFile].
	suffix = 'htm' | (suffix = 'html')
		ifTrue: [labels add: 'open in browser'.
			selectors add: #openInBrowser].
	suffix = 'zip' | (suffix = '*')
		ifTrue: [labels add: 'open archive viewer'.
			selectors add: #openArchiveViewer].
	suffix = '*'
		ifTrue: [labels addAll: #('generate HTML' ).
			lines add: labels size - 1.
			selectors addAll: #(#renderFile )].
	suffix = CRDictionary fileNameSuffix
		ifTrue: [labels add: 'load Genie Gesture Dictionary'.
			selectors add: #loadCRDictionary].
	suffix = CRDisplayProperties fileNameSuffix
		ifTrue: [labels add: 'load Genie Display Properties'.
			selectors add: #loadCRDisplayProperties].
	^ Array
		with: labels
		with: lines
		with: selectors! !

!FileList methodsFor: 'menu messages' stamp: 'yo 2/19/2004 22:53'!
installTTFStyle

	TTCFont newTextStyleFromTTFile: (directory fullPathFor: fileName).
! !

